import os
import numpy as np
import numba as nb
import casatools
import time

from scipy.optimize import curve_fit

ms = casatools.ms()
tb = casatools.table()
me = casatools.measures()
qa = casatools.quanta()
ia = casatools.image()
im = casatools.imager()

@nb.njit(cache=True)
def populate_grid(uvw, stokesI, uvgrid, uvgrid_npt, deltau, deltav, npix):
    for ii in range(len(uvw[0])):
        uidx = int(np.round(uvw[0][ii]//deltau + npix//2))
        vidx = int(np.round(uvw[1][ii]//deltav + npix//2))

        uvgrid[uidx, vidx] += stokesI[ii]
        uvgrid_npt[uidx, vidx] += 1

    return uvgrid, uvgrid_npt


def mad(inpdat):
    """
    Calculate the STD via MAD for the input data

    Inputs:
    inpdat      Input numpy array

    Returns:
    std         Calculate the std via mad
    """

    med = np.median(inpdat)
    mad = np.median(np.abs(inpdat - med))

    return 1.4826 * mad


def msuvbin_to_uvgrid(ms, npix):
    tb.open(ms)
    uvw = tb.getcol('UVW')
    data = tb.getcol('DATA')
    tb.close()

    umin, umax = np.min(uvw[0]), np.max(uvw[0])
    vmin, vmax = np.min(uvw[1]), np.max(uvw[1])
    wmin, wmax = np.min(uvw[2]), np.max(uvw[2])

    uvals = np.linspace(umin, umax, npix)
    vvals = np.linspace(vmin, vmax, npix)

    deltau = uvals[1] - uvals[0]
    deltav = vvals[1] - vvals[0]

    uvgrid = np.zeros((npix, npix), dtype=np.complex128)
    uvgrid_npt = np.zeros((npix, npix), dtype=int)

    stokesI = 0.5*(data[0] + data[1])
    stokesI = np.squeeze(stokesI)

    uvgrid, uvgrid_npt = populate_grid(uvw, stokesI, uvgrid, uvgrid_npt, deltau, deltav, npix)
    # Average per uv cell
    idx = np.where(uvgrid_npt != 0)
    uvgrid[idx] = uvgrid[idx]/uvgrid_npt[idx]

    return uvgrid, uvgrid_npt


def resid_function(x, a, b, c, d):
    return a*x**3 + b*x**2 + c*x + d


def fit_radial_profile(xvals,  yvals, ystd, deg=3, clip_sigma=3):
    """
    Fit the radial profile with a polynomial
    """

    print(f"deg {deg}, clip_sigma {clip_sigma}")
    print(f"xvals shape {xvals.shape}, yvals shape {yvals.shape}")
    print(f"non-zero x {np.count_nonzero(xvals)}, non-zero y {np.count_nonzero(yvals)}")

    idx = np.where(yvals != 0)
    xvals = xvals[idx]
    yvals = yvals[idx]
    ystd = ystd[idx]
    print(f"xvals {xvals}, yvals {yvals}")

    # Fit the radial profile with a polynomial
    pfit, pcov = curve_fit(resid_function, xvals, yvals, sigma=ystd, p0=[1, 1, 1, 1])

    #pfit = np.polyfit(xvals, yvals, deg=deg)
    yfit = np.polyval(pfit, np.linspace(xvals.min(), xvals.max(), 100))

    has_converged = False
    while not has_converged:
        # Outlier rejection while fitting
        resid = yvals - np.polyval(pfit, xvals)
        resid_mad = mad(resid)
        resid_med = np.median(resid)
        idx = np.where((resid > resid_med - clip_sigma*resid_mad) & (resid < resid_med + clip_sigma*resid_mad))

        if len(idx[0]) == len(xvals):
            has_converged = True
            continue

        xvals = xvals[idx]
        yvals = yvals[idx]
        ystd = ystd[idx]

        #pfit = np.polyfit(xvals, yvals, deg=deg)

        pfit, pcov = curve_fit(resid_function, xvals, yvals, sigma=ystd, p0=[1, 1, 1, 1])
        yfit = np.polyval(pfit, np.linspace(xvals.min(), xvals.max(), 100))

    #import matplotlib.pyplot as plt

    #fig, ax = plt.subplots()
    #ax.plot(xvals, yvals, '-o', label='Data')
    #ax.plot(np.linspace(xvals.min(), xvals.max(), 100), yfit, '--', label='Fit')
    #ax.set_xlabel('Radius')
    #ax.set_ylabel('Intensity')
    #ax.legend()
    #plt.tight_layout()
    #plt.show()

    return pfit


def calc_radial_profile_ann(uvgrid, deltau, deltav, doplot=False):
    """
    Calculate the annular average of the uvgrid for every radius,
    and fit the 1D radial profile with a polynomial.
    """

    nbin = 20

    npixx, npixy = uvgrid.shape[0], uvgrid.shape[1]
    cx, cy = npixx//2, npixy//2

    print(f"npixx {npixx}, npixy {npixy}, centerx {cx} centery {cy}")

    # Generate grid of radii
    x = np.arange(npixx) - cx
    y = np.arange(npixy) - cy
    rad = np.sqrt(x**2 + y**2).astype(int)

    yy, xx = np.meshgrid(x, y)
    radgrid = np.sqrt(xx**2 + yy**2).astype(int)

    print(f"x shape {x.shape}, y shape {y.shape}, rad shape {rad.shape}")
    print(f"x min {x.min()}, x max {x.max()}")
    print(f"y min {y.min()}, y max {y.max()}")
    print(f"rad min {rad.min()}, rad max {rad.max()}")

    # Create log-spaced annuli to account for reducing UV coverage with radius
    # Minimum annulus is 5px
    annuli = np.logspace(np.log10(5), np.log10(rad.max()), nbin)
    annuli = np.round(annuli).astype(int)

    radial_mean = np.zeros(nbin)
    radial_mad = np.zeros(nbin)

    ann_min = 0
    ann_max = 0
    for idx, ann in enumerate(annuli):
        ann_max = ann_min + ann

        ridx = np.where((radgrid >= ann_min) & (radgrid < ann_max))
        uvgrid_sel = uvgrid[ridx]
        uvgrid_sel = uvgrid_sel[np.abs(uvgrid_sel) != 0]

        print(f"ann_min {ann_min} ann_max {ann_max}")
        print(f"len(ridx) {len(ridx[0])}")
        print(f"len(uvgrid_sel) {len(uvgrid_sel)}")

        if len(uvgrid_sel) == 0:
            radial_mean[idx] = 0
            radial_mad[idx] = 0.
        else:
            radial_mean[idx] = np.mean(uvgrid_sel)
            radial_mad[idx] = mad(uvgrid_sel)

        print(f"radial_mean {radial_mean[idx]} radial_mad {radial_mad[idx]}")
        print('-------------------------------')

        ann_min += ann

    print(f"pre flag non-zero points {np.count_nonzero(uvgrid)}")
    # Flag the data
    ann_min = 0
    ann_max = 0
    for idx, ann in enumerate(annuli):
        ann_max = ann_min + ann
        #print(f"ann_min {ann_min} ann_max {ann_max}")

        ridx = np.where((radgrid >= ann_min) & (radgrid < ann_max))
        uvgrid_sel = uvgrid[ridx]

        flagidx = np.where(np.abs(uvgrid_sel) > radial_mean[idx] + 5*radial_mad[idx])
        uvgrid_sel[flagidx] = 0

    print(f"post flag non-zero points {np.count_nonzero(uvgrid)}")
    return uvgrid, radial_mean, radial_mad, annuli


def calc_radial_profile_pix(uvgrid, deltau, deltav, doplot=False):
    """
    Calculate the radial per-pixel average of the uvgrid for every radius,
    and fit the 1D radial profile with a polynomial.
    """

    npixx, npixy = uvgrid.shape[0], uvgrid.shape[1]
    cx, cy = npixx//2, npixy//2

    print(f"npixx {npixx}, npixy {npixy}, centerx {cx} centery {cy}")

    # Generate radial values from 0 to max
    x = np.arange(cx)
    y = np.arange(cy)
    rad = np.sqrt(x**2 + y**2).astype(int)

    print(f"x shape {x.shape}, y shape {y.shape}, rad shape {rad.shape}")
    print(f"rad [0] {rad[0]}, rad [-1] {rad[-1]}")
    print(f"rad {rad}")

    # Generate grid of radii
    x = np.arange(npixx) - cx
    y = np.arange(npixy) - cy
    yy, xx = np.meshgrid(x, y)
    radgrid = np.sqrt(xx**2 + yy**2).astype(int)

    radial_mean = np.zeros(np.max([cx, cy]))
    radial_mad = np.zeros(np.max([cx, cy]))

    for idx, rr in enumerate(rad):
        if idx == len(rad) - 1:
            ridx = np.where((radgrid >= rad[idx]))
        else:
            ridx = np.where((radgrid > rad[idx]) & (radgrid <= rad[idx+1]))

        uvgrid_sel = uvgrid[ridx]
        uvgrid_sel = uvgrid_sel[np.abs(uvgrid_sel) != 0]

        if len(uvgrid_sel) == 0:
            radial_mean[idx] = 0
            radial_mad[idx] = 0.
        else:
            radial_mean[idx] = np.mean(uvgrid_sel)
            radial_mad[idx] = mad(uvgrid_sel)

    return uvgrid, radial_mean, radial_mad, rad*deltau


@nb.njit(cache=True)
def apply_flags(dat, flg, uvw, radial_mean, radial_mad, annuli):
    """
    Apply flags based on the radial profile to the input data column
    """

    print(f"dat.shape {dat.shape}, flg.shape {flg.shape}, uvw.shape {uvw.shape}")

    nrow = uvw.shape[1]

    for rr in range(nrow):
        uvlen = np.sqrt(uvw[0][rr]**2 + uvw[1][rr]**2)
        annidx = np.searchsorted(annuli, uvlen)

        if np.abs(dat[..., rr]) > radial_mean[annidx] + 5*radial_mad[annidx]:
            flg[..., rr] = True

    return flg


def flagViaBin_radial(binnedvis="", improveImp=True, sigma=5):
    tb.open(binnedvis, nomodify=False)
    msuvbinkey = tb.getkeyword('MSUVBIN')

    # msuvbinkey.keys()
    # Out[5]: dict_keys(['csys', 'nchan', 'npol', 'numvis', 'nx', 'ny', 'sumweight'])

    # in radian
    dra = msuvbinkey['csys']['direction0']['cdelt'][0]
    ddec = msuvbinkey['csys']['direction0']['cdelt'][1]

    nx = msuvbinkey['nx']
    ny = msuvbinkey['ny']

    # in radian
    ra_extent = dra*nx
    dec_extent = ddec*ny

    # in Lambda
    deltau = 1./ra_extent
    deltav = 1./dec_extent

    npol = msuvbinkey['npol']
    nchan = msuvbinkey['nchan']

    npoints = min(nx, ny)

    print(f"nx {nx}, ny {ny}, npol {npol}, nchan {nchan}, npoints {npoints}")
    print(f"dra {dra} ddec {ddec} ra_extent {ra_extent} dec_extent {dec_extent} deltau {deltau} deltav {deltav}")

    dat = tb.getcol('UVW')
    print(f"Data shape {dat.shape}")

    for pol in range(npol):
        for chan in range(nchan):

            dat = tb.getcolslice('DATA', [pol, chan], [pol, chan], [1, 1])
            flg = tb.getcolslice('FLAG', [pol, chan], [pol, chan], [1, 1])
            wgt = tb.getcolslice('WEIGHT_SPECTRUM', [pol, chan], [pol, chan], [1, 1])
            uvw = tb.getcol('UVW')

            print(f"pol {pol}, chan {chan}")
            print(f"data shape {dat.shape}, flag shape {flg.shape}, weight shape {wgt.shape} uvw shape {uvw.shape}")

            if dat.size == 0 or flg.size == 0 or wgt.size == 0:
                print(f"Zero size array read. Skipping.")
                continue

            dat_grid = dat[0, 0, :].reshape([npoints, npoints])
            flg_grid = flg[0, 0, :].reshape([npoints, npoints])
            wgt_grid = wgt[0, 0, :].reshape([npoints, npoints])

            # Flag the data as necessary
            dat_grid = dat_grid * ~flg_grid

            # NOTE: The original algorithm calls for radial fitting - this just calculates the median
            # as a function of radius. Similar, not the same. Radial fitting with outlier rejection
            # may yield better results.
            uvgrid, radial_mean, radial_mad, annuli = calc_radial_profile_ann(dat_grid, deltau, deltav, doplot=True)
            #radial_fit = fit_radial_profile(annuli,  radial_mean, radial_mad, deg=2)

            import matplotlib.pyplot as plt
            from matplotlib.colors import LogNorm
            fig, ax = plt.subplots()
            ax.plot(annuli, radial_mean, '-o', label='data')
            #ax.plot(np.linspace(np.min(annuli), np.max(annuli), 1000), np.polyval(radial_fit, np.linspace(np.min(annuli), np.max(annuli), 1000)), '--', label='Fit')
            ax.fill_between(annuli, radial_mean-radial_mad, radial_mean+radial_mad, alpha=0.5)
            ax.set_xlabel('Radius (lambda)')
            ax.set_ylabel('Intensity (Jy)')
            ax.set_title('Radial Mean')
            ax.set_yscale('symlog')
            ax.legend()
            plt.tight_layout()

            plt.savefig('radprof.jpg', bbox_inches='tight')

            fig, ax = plt.subplots(1, 1)
            ax.imshow(np.abs(uvgrid), origin='lower', norm=LogNorm(vmin=1e-12, vmax=1))
            ax.set_title('UV grid')
            plt.tight_layout()
            #plt.savefig('uvgrid.jpg', bbox_inches='tight')
            plt.show()

            print(f"dat_grid shape {dat_grid.shape}, flg_grid shape {flg_grid.shape}, wgt_grid shape {wgt_grid.shape}")

            # Do the flagging and write back

            flg_new = apply_flags(dat, flg, uvw, radial_mean, radial_mad, annuli)

            print("Old flagged points: ", np.count_nonzero(flg))
            print("Flagged points: ", np.count_nonzero(flg_new))

            tb.putcolslice('DATA', dat, [pol, chan], [pol, chan])
            tb.putcolslice('FLAG', flg, [pol, chan], [pol, chan])
            tb.putcolslice('WEIGHT_SPECTRUM', wgt, [pol, chan], [pol, chan])

        tb.clearlocks()
        tb.done()


# @saveparams2last(multibackup=True)
def msuvbinflag(
        ######## binned ms
        binnedvis, #= 'thegridded_from_msuvbin.ms',
        ######## flagging algorithm option
        method, #= 'radial','RegionalMean','Gradient'
        sigma, #=0.5
        sizeRegion, #=20
        radius, #=1.0
        ignorPoint #= True
) -> None:

    if method == 'radial':
        tic = time.time()
        flagViaBin_radial(binnedvis=binnedvis, sigma=5)
        toc = time.time()
        print("MSuvbinflag radial method Running time :", toc - tic)
    elif method == 'regionalMean':
        print("No bueno")
        #tic = time.time()
        #flagViaBin_regionalMean(binnedvis=binnedvis, sizeRegion=20, sigma=5,ignorPoint=True)
        #toc = time.time()
        #print("MSuvbinflag other method Running time:", toc - tic)
    elif method == 'gradient':
        print("No bueno")
        #tic = time.time()
        #flagViaBin_gradient(binnedvis=binnedvis,  radius=1, sigma=0.3)
        #toc = time.time()
        #print("MSuvbinflag pead method Running time:", toc - tic)
    elif method == 'median':
        print("No bueno")
        #tic = time.time()
        #flagViaBin_median(binnedvis=binnedvis,  sigma=5)
        #toc = time.time()
        #print("MSuvbinflag median method Running time:", toc - tic)
    else:
        print("No bueno")
    return
